importClass(Packages.java.security.KeyStore);
importClass(Packages.java.security.Signature);
importClass(Packages.org.apache.axis.encoding.Base64);

// These variables are used by E4X to access nodes prefixed with a namespace. These namespaces are defined
// in the SOAP Response that is parsed by E4X
var sd = new Namespace("http://www.ca.com/UnicenterServicePlus/ServiceDesk");

var CAServiceDeskWS = WSUtil.extend({

  LOG_SOURCE: "caservicedeskws.js: ",

// CA Service Desk Web Service Constants
  ATTRIBUTES_STATUS_LIST: new Array("code", "sym"),
  ATTRIBUTES_CONTACT_TYPE_LIST: new Array("id", "sym", "delete_flag"),
  ATTRIBUTES_CNT_OBJECT: new Array("alt_phone", "type", "email_address", "fax_phone", "first_name", "delete_flag", "last_name", "beeper_phone", "pemail_address", "phone_number", "userid", "id"),
  ATTRIBUTES_GRPMEM_OBJECT: new Array("group", "member", "id"),

  // This list should contain all those values added to the Addtional Tokens collection in the APS that are available for EVERY response.
  DEFAULT_TOKENS: new Array("incident_id", "object_handle", "target", "targetpwd"),
  CA_SOAP_VERSION: "1_1",
  MAX_ROWS: "100",

  SERVICE_DESK_WSDL_DEFN: "integrationservices/caservicedesk/soap/CA-Service-Desk-soapui-project.xml",

  /**
   * Constructor
   * @param endPoint SOAP EndPoint for CA Service Desk Manager
   * @param sid session id for the webservices session
   */
  init: function(endPoint, sid)
  {
    this._super(); 
    this.endPoint = endPoint;
    this.sid = sid;
  },
  
  /******************************************************************************/
  /*         Start of Webservice Call methods for CA Service Desk               */
  /******************************************************************************/

  /******************************************************************************/
  /*                           Public Methods                                   */
  /******************************************************************************/

  /**
   * Creates a new activity log entry for the selected incident
   * @param objectHandle id of the incident to annotate
   * @param creator id of the user updating the incident log
   * @param description message to annotate
   * @param logType type of log entry to create
   * @param internal true if this log is internal only, false otherwise
   */
  annotate: function(objectHandle, creator, description, logType, internal)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "createActivityLog", "createActivityLog", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("creator", creator);
    msg.setParameter("objectHandle", objectHandle);
    msg.setParameter("description", description);
    msg.setParameter("logType", !this.isEmpty(logType) ? logType : "LOG");
    msg.setParameter("timeSpent", "0");
    msg.setParameter("internal", !this.isEmpty(internal) ? internal : "false");
  
    // Post the message to the SOAP endpoint
    return this.sendReceive(this.endPoint, msg);
  },
  
  /**
   * Update the incident with the new status and assignee (if applicable)
   * @param incidentId id of the incident to update
   * @param status new status of the incident
   * @param additionalTokens Collection of attributes to be added to the SOAP Request
   * @param objectHandle ID of the object being updated
   */
  updateTicket: function(incidentId, status, additionalTokens, objectHandle)
  {
    var attrVals = "<string>status</string><string>" + status.toUpperCase() + "</string>";
    
    attrVals += this.getAttributeValues(additionalTokens);
  
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "updateObject", "updateObject", this.CA_SOAP_VERSION);
    
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("objectHandle", objectHandle);
    msg.setParameter("attrVals", attrVals, false);
    msg.setParameter("attributes", "", false);
  
    // Post the message to the SOAP endpoint
    return this.sendReceive(this.endPoint, msg);
  },

  /**
   * Create a new ticket based on the role of the user creating the ticket
   * @param customerHandle id of the user creating the ticket
   * @param description summary of the ticket being created
   */
  createQuickTicket: function(customerHandle, description)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "createQuickTicket", "createQuickTicket", this.CA_SOAP_VERSION);
    
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("customerHandle", customerHandle);
    msg.setParameter("description", description);
  
    // Post the message to the SOAP endpoint
    return this.sendReceive(this.endPoint, msg);
  },

  /**
   * Translates the ticket type into a valid factory of the ticket
   * @param ticketType String containing the ticket type
   * @return ticketFactory for the ticket type.
   */
  getTicketFactory: function(ticketType)
  {
    if ("Incident".equalsIgnoreCase(ticketType))
    {
      return "in";
    }
    if ("Request".equalsIgnoreCase(ticketType))
    {
      return "cr";
    }
    if ("Problem".equalsIgnoreCase(ticketType))
    {
      return "pr";
    }
    return ticketType;
  },

  /**
   * Retrieves the response options that will be presented to the user on the notification
   * @param objectHandle identifies the handle for a ticket or status
   * @param ticketFactory Identifies the factory of the ticket. Valid values are only "cr" for Request, 
   *        "in" for Incident, "pr" for Problem, "chg" for Change Order and "iss" for Issue.
   * @return array of valid responses for the current ticket status in the form key=value.
   */
  getResponseOptions: function(objectHandle, ticketFactory)
  {
    // Retrieve the transitions for the object/ticket factory
    var response = this.getValidTransitions(objectHandle, ticketFactory);
    
    // Retrieve the map of status defined for this SDM instance
    var statusMap = this.getStatusMap(ticketFactory);
  
    var validResponses = [];
    var index = 0;
  
    this.log.debug("Valid Transitions retrieved: " + response.UDSObject.length());
    
    // Add Ignore as the first option if enabled for this integration
    if (ADD_IGNORE_RESPONSE)
    {
        validResponses[index++] = "IGN=Ignore";
    }
    
    // Iterate the returned transitions and extract the valid response options for
    // the valid next statuses defined in the transitions.
    for (var i=0; i<response.UDSObject.length(); i++)
    {
      var attributes = response.UDSObject[i].Attributes;
      for (var j=0; j<attributes.Attribute.length(); j++)
      {
        var attribute = attributes.Attribute[j];
        if ("new_status".equalsIgnoreCase(attribute.AttrName))
        {
          var key = attribute.AttrValue;
          this.log.debug("new_status=" + key)
          if (key != null)
          {
            this.log.debug("Map Value: " + statusMap.get(key.toUpperCase()));
            validResponses[index++] = key + "=" + statusMap.get(key.toUpperCase());
          }
        }
      }
    }
    validResponses[index++] = "Annotate=Annotate";
    this.log.debug("ValidResponses: " + validResponses.toString());
    return validResponses;
  },
  
  /**
   * Logs the user into CA Service Desk
   * @param userId ID of user to login for this session
   * @param password String password for the user
   * @return session id for further requests
   */
  login: function(userId, password)
  {
    if (MANAGED_LOGIN)
    {
      this.loginServiceManaged(userId);
    }
    else
    {
      var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "login", "login", this.CA_SOAP_VERSION);
  
      // Set the parameter values
      msg.setParameter("username", userId);
      msg.setParameter("password", password);
  
      // Post the message to the SOAP endpoint
      var response = this.sendReceive(this.endPoint, msg);
      this.sid = response.sd::loginResponse.loginReturn;
    
      if (this.sid <= 0)
      {
        throw new Exception("Unable to create WebService Session");
      }
    }
  },

  /**
   * Session ID for the current WS Session.
   * If the session is not we retrive it by logging in.
   */
  getSessionId: function() 
  {
    if (this.sid == null || this.sid == "")
    {
      throw new Exception("Unable to create WebService Session, you must login first");
    }

    return this.sid;
  },

  /******************************************************************************/
  /*                           Private Methods                                  */
  /******************************************************************************/

  /**
   * Checks if the given value is null or empty string
   * @param value parameter to check
   * @return true if the value is empty, false otherwise
   */
  isEmpty: function(value)
  {
    if (value == null || "".equalsIgnoreCase(value) || value == "")
    {
      return true;
    }
    return false;
  },
  
  /**
   * Logs the user into CA Service Desk using the CA Service Desk Manager Certificate
   * @param userId ID of user that will be impersonated if impersonation is enabled
   * @return session id for further requests
   */
  loginServiceManaged: function(userId)
  {
    // Get the appropriate Keystore instance to load the .p12 file
    var ks = KeyStore.getInstance("PKCS12");
    
    // Create a password to be used when extracting the private key. Same as the access policy.
    var privateKeyPassword = ACCESS_POLICY.split("");
    
    // Load the certificate
    var dir = ACCESS_POLICY_FILE_PATH.lastIndexOf("/") != -1 ? ACCESS_POLICY_FILE_PATH : ACCESS_POLICY_FILE_PATH + "/";
    ks.load(new FileInputStream(dir + ACCESS_POLICY + ".p12" ), privateKeyPassword);
    
    // Extracting the private key, the first parameter is the alias associated 
    // with the key which is "servicedesk <accessPolicy>"(all lower case); 
    // the second parameter is the password to extract the key (this defaults to 
    // the Access Policy name as well ... when using pdm_pki utility)
    var key = ks.getKey("servicedesk " + ACCESS_POLICY.toLowerCase(), privateKeyPassword);
    
    // Creating an instance of the signature class to create a digital signature 
    // of the private key...must use the SHA1withRSA algorithm
    var s = Signature.getInstance("SHA1withRSA"); 
    
    // Initializing the signature with the private key
    s.initSign(key);
        
    // Updating the signature with the access policy
    var b = new java.lang.String(ACCESS_POLICY).getBytes();
    s.update(b, 0, b.length);
        
    // Creating the digital signature
    var sig = s.sign();
        
    // Converting the signature to BASE64 text format to pass into loginServiceManaged
    var encryption = Base64.encode(sig);		
  
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "loginServiceManaged", "loginServiceManaged", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("policy", ACCESS_POLICY);
    msg.setParameter("encrypted_policy", encryption);
  
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    this.sid = response.sd::loginServiceManagedResponse.loginServiceManagedReturn;
    
    if (this.sid <= 0)
    {
      throw new Exception("Unable to create WebService Session");
    }
  
    // If impersonation is enabled, impersonate the selected user.  
    if (ENABLE_IMPERSONATE)
    {
      this.impersonate(userId);
    }
  },
  
  /**
   * Logs the user into CA Service Desk impersonating the requested user
   * @param userId name of user to impersonate  
   * @return session id for further requests
   */
  impersonate: function(userId)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "impersonate", "impersonate", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.sid);
    msg.setParameter("userid", userId);
  
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    this.sid = response.sd::impersonateResponse.impersonateReturn;
    
    if (this.sid <= 0)
    {
      throw new Exception("Unable to create WebService Session");
    }
  },
  
  /**
   * Logs the current user session out of CA Service Desk
   */
  logout: function()
  {
    try
    {
      if (this.getSessionId() != null)
      {
        var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "logout", "logout", this.CA_SOAP_VERSION);
    
        // Set the parameter values
        msg.setParameter("sid", this.getSessionId());
      
        // Post the message to the SOAP endpoint
        this.sendReceive(this.endPoint, msg);
      }
    }
    catch (e)
    {
      this.log.warn("Unable to logout of the webservices session: " + e.toString());
    }
  },
  
  /** 
   * Retrieves the requested information from Service Desk  
   * @param objectType type of Object to retrieve 
   * @param whereClause filter that limits the results retrieved 
   * @param maxRows specifies the maximum number of records that will be returned 
   * @param attributes array of attributes that will be returned in the resultset 
   * @return XML fragment containing the requested data 
   */  
  doSelect: function(objectType, whereClause, maxRows, attributes)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "doSelect", "doSelect", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("objectType", objectType);
    msg.setParameter("whereClause", whereClause);
    msg.setParameter("maxRows", maxRows);
    msg.setParameter("attributes", this.createListElements("string", attributes), false);
    
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    response = this.formatStringForE4X(response.sd::doSelectResponse.doSelectReturn);
    return new XML(response);
  },
  
  /**
   * Retrieves a list of objectType from Service Desk
   * @param objectType type of Object to retrieve
   * @param whereClause similar to SQL WHERE statement to limit objectTypes included in list
   * @return XML fragment that consists of the doQueryReturn element and its children
   */
  doQuery: function(objectType, whereClause)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "doQuery", "doQuery", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("objectType", objectType);
    msg.setParameter("whereClause", whereClause);
    
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    // No need to format for E4X as doQuery responses should not have XML declarations.
    return new XML(response.sd::doQueryResponse.doQueryReturn);
  },
  
  /**
   * Retrieves a list of objectType from Service Desk
   * @param handles array of list handle reference ids
   * @return XML fragment that consists of the doQueryReturn element and its children
   */
  freeListHandles: function(handles)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "freeListHandles", "freeListHandles", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("handles", this.createListElements("integer", handles), false);
    
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
  },
  
  /**
   * Gets the values of the list created by a doQuery call
   * @param listHandle listHandle returned from doQuery
   * @param startIndex index within the list to start
   * @param endIndex index within the list to end
   * @param attributes attributes of the objectType to return
   * @return details
   */
  getListValues: function(listHandle, startIndex, endIndex, attributeNames)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "getListValues", "getListValues", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("listHandle", listHandle);
    msg.setParameter("startIndex", startIndex);
    msg.setParameter("endIndex", endIndex);
    msg.setParameter("attributeNames", this.createListElements("string", attributeNames), false);
    
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    response = this.formatStringForE4X(response.sd::getListValuesResponse.getListValuesReturn);
    return new XML(response);
  },
  
  /**
   * Retrieves the GUID for the specified user
   * @param userId name of the user to lookup
   * @return user GUID
   */ 
  getHandleForUserId: function(userId)
  {
    if (userId != null)
    {
      var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "getHandleForUserid", "getHandleForUserid", this.CA_SOAP_VERSION);
      
      // Set the parameter values
      msg.setParameter("sid", this.getSessionId());
      msg.setParameter("userID", userId);
   
      // Post the message to the SOAP endpoint
      var response = this.sendReceive(this.endPoint, msg);
      return response.sd::getHandleForUseridResponse.getHandleForUseridReturn;
    }
    return null;
  },

  /**
   * Retrieves the contact details for the specified contact ID
   * @param contactId id of the contact to lookup
   * @return contact information
   */ 
  getContact: function(contactId)
  {
    if (contactId != null)
    {
      var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "getContact", "getContact", this.CA_SOAP_VERSION);
      
      // Set the parameter values
      msg.setParameter("sid", this.getSessionId());
      msg.setParameter("contactId", contactId);
   
      // Post the message to the SOAP endpoint
      var response = this.sendReceive(this.endPoint, msg);
      return response.sd::getContactResponse.getContactReturn;
    }
    return null;
  },

  /**
   * Retrieves the related list for the object handle
   * @param objectHandle id of the object the list relates to
   * @param listName name of the list that is being retrieved
   * @return list information to allow retrieval of the values
   */ 
  getRelatedList: function(objectHandle, listName)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "getRelatedList", "getRelatedList", this.CA_SOAP_VERSION);
    
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("objectHandle", objectHandle);
    msg.setParameter("listName", listName);
 
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    return response.sd::getRelatedListResponse.getRelatedListReturn;
  },
  
  /**
   * Retrieves the transitions that have been defined for the workflow
   * @param objectHandle identifies the handle for a ticket or status
   * @param ticketFactory Identifies the factory of the ticket. Valid values are only "cr" for Request, 
   *        "in" for Incident, "pr" for Problem, "chg" for Change Order and "iss" for Issue.
   * @return all of the possible transition values for the status of a ticket
   */ 
  getValidTransitions: function(objectHandle, ticketFactory)
  {
    var msg = new SOAPMessage(this.SERVICE_DESK_WSDL_DEFN, "getValidTransitions", "getValidTransitions", this.CA_SOAP_VERSION);
  
    // Set the parameter values
    msg.setParameter("sid", this.getSessionId());
    msg.setParameter("handle", objectHandle);
    msg.setParameter("ticketFactory", ticketFactory);
   
    // Post the message to the SOAP endpoint
    var response = this.sendReceive(this.endPoint, msg);
    response = this.formatStringForE4X(response.sd::getValidTransitionsResponse.getValidTransitionsReturn);
    return new XML(response);
  },
  
  /**
   * Retrieves the map of defined status for the system and their long name
   * @param objectType name of the object containing the status translations
   * @return a map of status code/status description for use in response options
   */
  getStatusMap: function(objectType)
  {
    if ("chg".equalsIgnoreCase(objectType))
    {
      objectType = "chgstat";
    }
    else if ("iss".equalsIgnoreCase(objectType))
    {
      objectType = "issstat";
    }
    else
    {
      objectType = "crs";
    }
  
    var statusList = this.doSelect(objectType, "code like '%'", this.MAX_ROWS, this.ATTRIBUTES_STATUS_LIST);
    var statusMap = new HashMap();
    
    this.log.debug("Size of StatusList retrieved: " + statusList.UDSObject.length());
    
    for (var i=0; i<statusList.UDSObject.length(); i++)
    {
      var attributes = statusList.UDSObject[i].Attributes;
      var key = null;
      var value = null;
      for (var j=0; j<attributes.Attribute.length(); j++)
      {
        var attribute = attributes.Attribute[j];
        this.log.debug("Processing attribute: " + attribute.AttrName);
        if ("code".equalsIgnoreCase(attribute.AttrName))
        {
          key = attribute.AttrValue;
          this.log.debug("key=" + key);
        }
        if ("sym".equalsIgnoreCase(attribute.AttrName))
        {
          value = attribute.AttrValue;
          this.log.debug("value=" + value);
        }
      }
      this.log.debug("Adding " + key + " to Map with value " + value);
      statusMap.put(key.toUpperCase(), value);
    }
    return statusMap;
  },
  
  
  /**
   * Can be used to create any children elements that represent an list item. 
   * @param nodeName name to give the node the attribute will be enclosed in
   * @param arrayList array of attributes that will be added to the collection
   * @return String containining a formatted list
   */
  createListElements: function(nodeName, arrayList)
  {
    var listElements = "";
    for (var i=0; i<arrayList.length; i++)
    {
      listElements += "<" + nodeName + ">" + arrayList[i].toString().trim() + "</" + nodeName + ">";
    }
    return listElements;
  },
  
  /**
   * Extract any additional tokens from the map and build the attribute list
   * @param additionalTokens map of tokens to process
   */
  getAttributeValues: function(additionalTokens)
  {
    var attrVals = "";
  
    // Retrieve any custom values from the additionalTokens and add to the attribute list
    for (var it = additionalTokens.keySet().iterator(); it.hasNext();)
    {
      var key = it.next();
      var value = additionalTokens.get(key);
      var ignoreToken = false;
      
      // If the token is NOT in the default tokens, add it to the return value
      for (var j=0; j<this.DEFAULT_TOKENS.length; j++)
      {
        if (this.DEFAULT_TOKENS[j].equalsIgnoreCase(key))
        {
          ignoreToken = true;
        }
      }
      
      if (!ignoreToken)
      {
        attrVals += "<string>" + key + "</string>";
        // Special Handling to convert responder name to a CA SDM ID.
        if ("assignee".equalsIgnoreCase(key) && !this.isEmpty(value))
        {
          value = this.getHandleForUserId(value);
        }
        attrVals += "<string>" + value + "</string>";
      }    
    }
    
    return attrVals;
  },

  /**
   * Call this function to extract contact records from CA SDM and create an array
   * of User objects which can be synchronized to xMatters 
   * @param objectType String containing the object that will be queried to retrieve the list values
   * @param whereClause filter to retrict the results
   * @param attributes array of attributes to be returned by the query
   * @param startIndex starting index in the list being processed
   * @param chunkSize number of records to process
   * @param reverse true if the list should be read from the end, false otherwise
   * @param listHandle if this list has already been retrieved, this is the handle for the list
   * @param listLength the number of items in the list
   */ 
  processListChunk: function(objectType, whereClause, attributes, startIndex, chunkSize, reverse, listHandle, listLength)
  {
    if (startIndex == null || chunkSize == null)
    {
      return this.processListValues(objectType, whereClause, attributes)
    }
    
    this.log.debug("Starting processListChunk");
  
    if (listHandle == null || listLength == null)
    {
      var doQueryReturn = this.doQuery(objectType, whereClause);
      this.log.debug("Finish doQuery");
      listLength = doQueryReturn.listLength;
      listHandle = doQueryReturn.listHandle;
    }
    var udsObjectListArray = new Array();
    
    // If the list is being read backwards, we need to recalculate the start and end indexes
    var endIndex = parseInt(startIndex) + parseInt(chunkSize);
    if ("true".equalsIgnoreCase(reverse))
    {
      var pageIndex = (startIndex / chunkSize) + 1;
      startIndex = listLength - (pageIndex * chunkSize);
      endIndex = (listLength - 1) - ((pageIndex - 1) * chunkSize);
      startIndex = startIndex < 0 ? 0 : startIndex;
      log.debug("startIndex: " + startIndex);
      log.debug("endIndex: " + endIndex);
    }
    else
    {
      endIndex = (listLength < endIndex ? listLength : endIndex) - 1;
    }
    udsObjectListArray[0] = this.getListValues(listHandle, startIndex, endIndex, attributes);
    this.log.debug("Finish getListValues");
    
    this.log.debug("Finished processListChunk");
    return udsObjectListArray;
  },

  /**
   * Call this function to extract contact records from CA SDM and create an array
   * of User objects which can be synchronized to xMatters 
   * @param objectType String containing the object that will be queried to retrieve the list values
   * @param whereClause filter to retrict the results
   * @param attributes array of attributes to be returned by the query
   */ 
  processListValues: function(objectType, whereClause, attributes)
  {
    this.log.debug("Starting processListValues");
  
    var doQueryReturn = this.doQuery(objectType, whereClause);
    this.log.debug("Finish doQuery");
    var listLength = doQueryReturn.listLength;
    var udsObjectListArray = new Array();
    
    this.log.debug("List Length: " + listLength);
    if(listLength < CHUNK_SIZE)
    {
      udsObjectListArray[0] = this.getListValues(doQueryReturn.listHandle, 0, listLength - 1, attributes);
      this.log.debug("Finish getListValues");
    }
    else
    {
      var chunks = Math.floor(listLength / CHUNK_SIZE);
      this.log.debug("********************* chunks = " + chunks + " *********************************");
      var listRemainder = (listLength % CHUNK_SIZE);
      this.log.debug("********************* listRemainder = " + listRemainder + " *********************************");
      var startIndex = 0;
      var endIndex = CHUNK_SIZE - 1;
      var i = 0
      while(i < chunks)
      {
        udsObjectListArray[i] = this.getListValues(doQueryReturn.listHandle, startIndex, endIndex, attributes);
        startIndex += CHUNK_SIZE;
        endIndex += CHUNK_SIZE;
        i++;
      }
      
      if(listRemainder > 0)
      {
        endIndex = endIndex - CHUNK_SIZE + listRemainder
        udsObjectListArray[i] = this.getListValues(doQueryReturn.listHandle, startIndex, endIndex, attributes);
      }
    }
    
    this.freeListHandles(new Array(doQueryReturn.listHandle));
    this.log.debug("Finished processListValues");
    return udsObjectListArray;
  },

  /**
   * Processes the UDSObjectList returned from the SOAP Request
   * @param udsObjectListArray collection of UDSObjects to process
   * @param fn JS function that will be called for the particular type
   * of processing that is required.
   */
  processUDSObjectList: function(udsObjectListArray, fn)
  {
    this.log.debug("Initializing Object Array for the objects from Service Desk SOAP Response");
    var array = new ArrayList();

    // Each element in the array is a list of records from CA Service Desk
    // The total list of contacts is broken into sections so that a single
    // web service call does not return all CA SDM records from the selected object
    for (var i=0; i<udsObjectListArray.length; i++)
    {
      this.log.debug("Processing UDSObjectListArray element: " + i);
      var udsObjectList = udsObjectListArray[i];
      // Each UDSObjectList is a list of UDSObjects which are Service Desk objects
      for (var j=0; j<udsObjectList.UDSObject.length(); j++)
      {
        this.log.debug("Processing UDSObject element: " + j + " of UDSObjectListArray element: " + i);
        array.add(fn(udsObjectList.UDSObject[j].Attributes, this));
      }
    }
    return array;
  },
  
  /**
   * At implementation time it was found that Mozilla's implementation of E4X
   * does not support XML declarations (<? ... ?>), therefore if they exist in
   * the payload returned from CA SDM, it needs to be stripped out.
   * http://www.xml.com/pub/a/2007/11/28/introducing-e4x.html?page=4   
   * 
   * This method takes a parameter:
   * 1. Ensures it is a string
   * 2. Removes any XML declarations
   * 3. returns a string which can be used to create E4X objects such as
   *    XML or XMLList.                         
   */
  formatStringForE4X: function(string)
  {
    this.log.debug("XML string that needs XML declarations removed: " + string);
    string = string.toString();
    string = string.replace(/<\?(.*?)\?>/g,'');
    this.log.debug("XML declarations should be removed: " + string);
    return string;
  },

  /******************************************************************************/
  /*    Interface to extract contact records from CA Service Desk               */
  /******************************************************************************/

  /**
   * Retrieve the id of the contact type
   * @param contactType name of the contact type to lookup
   * @return id of the requested contact type
   */
  getContactTypeId: function(contactType)
  {
    var contactTypeList = this.doSelect("ctp", "sym = '" + contactType + "'", this.MAX_ROWS, this.ATTRIBUTES_CONTACT_TYPE_LIST);
    this.log.debug("Number of contacts retrieved: " + contactTypeList.UDSObject.length());
    
    var id = null;
    for (var i=0; i<contactTypeList.UDSObject.length(); i++)
    {
      this.log.debug("Looking for id of contact record: " + i);
      var attributes = contactTypeList.UDSObject[i].Attributes;
      for (var j=0; j<attributes.Attribute.length(); j++)
      {
        var attribute = attributes.Attribute[j];
        this.log.debug("attribute: " + attribute.AttrName);
        if ("id".equalsIgnoreCase(attribute.AttrName))
        {
          id = attribute.AttrValue;
          this.log.debug("id=" + id);
        }
      }
    }
    return id;
  },


  /**
   * Helper function which, given a web service response list of contacts creates
   * the User objects.
   * @param attributes collection of attributes from the current UDSObject Contact 
   * @param context instance of the object context that this function belongs to
   * @return Populated User Object
   */ 
  createSyncUsers: function(attributes, context)
  {
    var userSync = new xMattersUserSync(syncList);
    var user = new User();
    this.log.debug("Creating xMatters sync user from CA SDM data");
    for (var i=0; i<attributes.Attribute.length(); i++)
    {
      var attribute = attributes.Attribute[i];
      this.log.debug("Found user attribute: " + attribute.AttrName + "=" + attribute.AttrValue);
      if ("alt_phone".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var phone = new VoiceDevice(); 
        phone.deviceName = "Home Phone"; 
        phone.number = attribute.AttrValue; 
        user.devices.add(phone);  
        this.log.debug("alt_phone=" + attribute.AttrName);
      }
      if ("email_address".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var email = new EmailDevice(); 
        email.address = attribute.AttrValue; 
        user.devices.add(email);  
        this.log.debug("email_address=" + attribute.AttrValue);
      }
      if ("fax_phone".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var phone = new VoiceDevice(); 
        phone.deviceName = "Other Phone"; 
        phone.number = attribute.AttrValue; 
        user.devices.add(phone);  
        this.log.debug("fax_phone=" + attribute.AttrValue);
      }
      if ("first_name".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        user.firstName = attribute.AttrValue;
        this.log.debug("user.firstName=" + user.firstName);
      }
      if ("delete_flag".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        user.active = attribute.AttrValue.equalsIgnoreCase("0") ? true : false;
        this.log.debug("user.active=" + user.active);
      }
      if ("last_name".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        user.lastName = attribute.AttrValue;
        this.log.debug("user.lastName=" + user.lastName);
      }
      if ("beeper_phone".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var textPhone = new TextPhoneDevice(); 
        textPhone.number = attribute.AttrValue; 
        user.devices.add(textPhone);  
        this.log.debug("beeper_phone=" + attribute.AttrValue);
      }
      if ("pemail_address".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var email = new EmailDevice(); 
        email.address = attribute.AttrValue;
        email.deviceName = "Pager Email"
        user.devices.add(email);  
        this.log.debug("pemail_address=" + attribute.AttrValue);
      }
      if ("phone_number".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        var phone = new VoiceDevice();  
        phone.number = attribute.AttrValue; 
        user.devices.add(phone);  
        this.log.debug("phone_number=" + attribute.AttrValue);
      }
      if ("userid".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        user.targetName = attribute.AttrValue;
        user.webLogin = attribute.AttrValue;
        this.log.debug("user.webLogin=" + user.webLogin);
      }
    }
    userSync.syncUser(user);
  },
  
  /**
   * Helper function which, given a web service response list of groups creates
   * the Group objects.
   * @param attributes collection of attributes from the current UDSObject Contact 
   * @param context instance of the object context that this function belongs to
   * @return Populated Group Object
   */ 
  createSyncGroups: function(attributes, context)
  {
    // Initialize the SOAP Helper and Logger objects
    var groupSync = new xMattersGroupSync(syncList);
    Group.prototype.id = "";

    var group = new Group();
    this.log.debug("Creating xMatters sync group from CA SDM data");
    for (var i=0; i<attributes.Attribute.length(); i++)
    {
      var attribute = attributes.Attribute[i];
      this.log.debug("Found group attribute: " + attribute.AttrName + "=" + attribute.AttrValue);
      if ("delete_flag".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        group.active = attribute.AttrValue.equalsIgnoreCase("0") ? true : false;
        this.log.debug("group.active=" + group.active);
      }
      if ("id".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        group.id = attribute.AttrValue;
        this.log.debug("group.id=" + group.id);
      }
      if ("last_name".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        group.name = attribute.AttrValue;
        group.team.groupName = group.name;
        group.coverage.groupName = group.name;
        this.log.debug("group.name=" + group.name);
      }
    }
   
    return group;
  },
  
  /**
   * Retrieves the user members from a grpmem record
   * @param attributes collection of attributes from the current UDSObject grpmem 
   * @return memberId
   */ 
  processGroupMembers: function(attributes)
  {
    var memberId = null;
    this.log.debug("Looking for member attribute");
    for (var i=0; i<attributes.Attribute.length(); i++)
    {
      var attribute = attributes.Attribute[i];
      this.log.debug("Found attribute: " + attribute.AttrName);
      
      if ("member".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        this.log.debug("memberid=" + attribute.AttrValue);
        memberId = attribute.AttrValue;
      }
    }
    return memberId;
  },

  /**
   * Retrieves the user members from a contact record
   * @param attributes collection of attributes from the current UDSObject contact
   * @return GroupMember
   */ 
  retrieveGroupMemberRecord: function(attributes)
  {
    var member = new GroupMember();
    this.log.debug("Looking for userid attribute");
    for (var i=0; i<attributes.Attribute.length(); i++)
    {
      var attribute = attributes.Attribute[i];
      this.log.debug("Found attribute: " + attribute.AttrName);
      
      if ("userid".equalsIgnoreCase(attribute.AttrName) && attribute.AttrValue != "")
      {
        member.targetName = attribute.AttrValue.toString();
        this.log.debug("user.webLogin=" + member.targetName);
      }
    }
    return member;
  },
});
